/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
import { TokenType, isGroup, isToken, isSpace, isTokenType, } from './tokens.js';
import { startPos } from './position.js';
import { ArrayList, atEnd, notAtEnd } from './list.js';
export function match(p, items, failure, context) {
    const r = p(new ArrayList(items, context));
    if (r === null)
        return failure;
    if (notAtEnd(skipSpace(r[1])))
        return failure;
    return r[0];
}
export const noItems = (c) => new ArrayList([], c);
export const fail = _i => null;
export function succeed(t) { return i => [t, i]; }
export const discard = i => [void 0, noItems(i.context)];
export const rest = i => [i.toArray(), noItems(i.context)];
export const end = i => atEnd(skipSpace(i)) ? [void 0, noItems(i.context)] : null;
export const pos = i => notAtEnd(i) ? [i.item.start, i] : null;
export const newline = i => {
    while (notAtEnd(i) && isTokenType(i.item, TokenType.SPACE))
        i = i.next;
    if (!notAtEnd(i) || !isTokenType(i.item, TokenType.NEWLINE))
        return null;
    return [i.item, i.next];
};
export function skipSpace(i) {
    while (notAtEnd(i) && isSpace(i.item))
        i = i.next;
    return i;
}
export function collectSpace(i, acc) {
    while (notAtEnd(i) && isSpace(i.item)) {
        acc.push(i.item);
        i = i.next;
    }
    return i;
}
export function withoutSpace(p) {
    return i => p(skipSpace(i));
}
export function not(p, v) {
    return i => p(i) === null ? [v, i] : null;
}
export function follows(p) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return [r[0], i];
    };
}
export function seq(...patterns) {
    return i => {
        for (const p of patterns) {
            const r = p(i);
            if (r === null)
                return null;
            i = r[1];
        }
        return [void 0, i];
    };
}
export function seqTuple(...patterns) {
    return i => {
        const rs = [];
        for (const p of patterns) {
            const r = p(i);
            if (r === null)
                return null;
            rs.push(r[0]);
            i = r[1];
        }
        return [rs, i];
    };
}
export function alt(...alts) {
    return i => {
        for (const a of alts) {
            const r = a(i);
            if (r !== null)
                return r;
        }
        return null;
    };
}
export function scope(pf) {
    return i => {
        const scope = Object.create(null);
        const r = pf(scope)(i);
        if (r === null)
            return null;
        return [scope, r[1]];
    };
}
export function value(pf) {
    return i => {
        const scope = Object.create(null);
        const r = pf(scope)(i);
        if (r === null)
            return null;
        return [scope.value, r[1]];
    };
}
export function bind(target, key, pattern) {
    return i => {
        const r = pattern(i);
        if (r === null)
            return null;
        target[key] = r[0];
        return r;
    };
}
export function exec(thunk) {
    return i => {
        thunk(i);
        return [void 0, i];
    };
}
export function map(p, f) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return [f(r[0]), r[1]];
    };
}
export function mapm(p, f) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return f(r[0])(r[1]);
    };
}
export function group(opener, items, options = {}) {
    return i => {
        var _a, _b;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        if (!isGroup(i.item))
            return null;
        if (i.item.open.text !== opener)
            return null;
        const r = items(new ArrayList(i.item.items, opener));
        if (r === null)
            return null;
        if (!atEnd(r[1]))
            return null;
        return [r[0], ((_b = options.advance) !== null && _b !== void 0 ? _b : true) ? i.next : i];
    };
}
export function atomString(text, options = {}) {
    return map(atom(text, options), _t => text);
}
export function atom(text, options = {}) {
    return i => {
        var _a, _b, _c;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        if (!isToken(i.item))
            return null;
        if (i.item.type !== ((_b = options.tokenType) !== null && _b !== void 0 ? _b : TokenType.ATOM))
            return null;
        if (text !== void 0 && i.item.text !== text)
            return null;
        return [i.item, ((_c = options.advance) !== null && _c !== void 0 ? _c : true) ? i.next : i];
    };
}
export function anything(options = {}) {
    return i => {
        var _a, _b;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        return [i.item, ((_b = options.advance) !== null && _b !== void 0 ? _b : true) ? i.next : i];
    };
}
export function upTo(p) {
    return i => {
        const acc = [];
        while (true) {
            const r = p(i);
            if (r !== null)
                return [acc, i];
            if (!notAtEnd(i))
                break;
            acc.push(i.item);
            i = i.next;
        }
        return null;
    };
}
export function separatedBy(itemPattern, separator) {
    return i => {
        const acc = [];
        if (end(i) !== null)
            return [acc, noItems(i.context)];
        while (true) {
            {
                const r = itemPattern(i);
                if (r === null)
                    return null;
                acc.push(r[0]);
                i = r[1];
            }
            {
                const r = separator(i);
                if (r === null) {
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                    return null;
                }
                i = r[1];
            }
        }
    };
}
export function separatedOrTerminatedBy(itemPattern, separator) {
    return i => {
        const acc = [];
        if (end(i) !== null)
            return [acc, noItems(i.context)];
        while (true) {
            {
                const r = itemPattern(i);
                if (r === null)
                    return null;
                acc.push(r[0]);
                i = r[1];
            }
            {
                const r = separator(i);
                if (r === null) {
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                    return null;
                }
                else {
                    i = r[1];
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                }
            }
        }
    };
}
export function repeat(p, options = {}) {
    return i => {
        var _a;
        const acc = [];
        let needSeparator = false;
        const finish = () => { var _a; return (acc.length < ((_a = options.min) !== null && _a !== void 0 ? _a : 0)) ? null : [acc, i]; };
        while (true) {
            if (acc.length == ((_a = options.max) !== null && _a !== void 0 ? _a : Infinity))
                return [acc, i];
            if (needSeparator) {
                if (options.separator) {
                    const r = options.separator(i);
                    if (r === null)
                        return finish();
                    i = r[1];
                }
            }
            else {
                needSeparator = true;
            }
            const r = p(i);
            if (r === null)
                return finish();
            acc.push(r[0]);
            i = r[1];
        }
    };
}
export function option(p) {
    return repeat(p, { max: 1 });
}
//---------------------------------------------------------------------------
// Search-and-replace over Item
export function replace(items, outerContext, p, f, end = items.length > 0 ? items[items.length - 1].end : startPos(null)) {
    const walkItems = (items, end, context) => {
        let i = new ArrayList(items, context);
        const acc = [];
        while (notAtEnd(i = collectSpace(i, acc))) {
            const r = p(i);
            if (r !== null) {
                acc.push(...f(r[0], notAtEnd(i) ? i.item.start : end, notAtEnd(r[1]) ? r[1].item.start : end));
                i = r[1];
            }
            else if (isToken(i.item)) {
                acc.push(i.item);
                i = i.next;
            }
            else {
                acc.push(Object.assign(Object.assign({}, i.item), { items: walkItems(i.item.items, i.item.end, i.item.open.text) }));
                i = i.next;
            }
        }
        return acc;
    };
    return walkItems(items, end, outerContext);
}
//# sourceMappingURL=matcher.js.map