/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2023-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>

const __SYNDICATE__ = Syndicate;
const currentSyndicateTarget = () => Syndicate.Dataspace.local;

(() => {
  async function translateScripts() {
    // SchemaReady is defined on the global object in preserves-schema-browser.js
    // (https://gitlab.com/preserves/preserves/-/blob/f2e3dd93c85d6d7e8d848fce6b8e6efefdf4af81/implementations/javascript/packages/schema/preserves-schema-browser.js#L6)
    // but if we only have Preserves at the time this code executes, and not PreservesSchema,
    // then there will be no SchemaReady variable, and we are in "use strict" mode because of
    // the way the codebase is bundled, so we must check if it exists on globalThis rather than
    // just expecting it to be a global variable with value `undefined`.
    if (globalThis.SchemaReady) await globalThis.SchemaReady;

    const syndicateScripts =
          Array.from(document.getElementsByTagName('script'))
          .filter(s => (s.type === 'text/javascript+syndicate' ||
                        s.type === 'syndicate' ||
                        s.type === 'module+syndicate'));

    for (const script of syndicateScripts) {
      const isModule = script.type === 'module+syndicate';

      const sourceUrl = script.src || script.getAttribute('data-src') || false;
      const sourceCode = sourceUrl ? await (await fetch(sourceUrl)).text() : script.innerHTML;

      const compilationResult = SyndicateCompiler.compile({
        name: sourceUrl || script.id || 'anonymous-script-tag',
        source: sourceCode,
        module: isModule ? 'es6' : 'none',
        runtime: isModule ? void 0 : 'Syndicate',
        emitError: console.error,
      });

      const sourceMap = { ... compilationResult.map };
      sourceMap.sourcesContent = [sourceCode];
      const formattedSourceMap = '\n//# sourceMappingURL=data:application/json;base64,' +
            Syndicate.Bytes.from(JSON.stringify(sourceMap)).toBase64();
      const finalOutput = compilationResult.text + formattedSourceMap;

      // console.log(finalOutput);

      const replacement = document.createElement('script');
      replacement.text = finalOutput;
      if (isModule) replacement.type = 'module';
      script.parentNode.replaceChild(replacement, script);
    }
  }

  window.addEventListener('DOMContentLoaded', translateScripts);
})();
